#!/usr/local/bin/perl

# Copyright (c) 2003 by Jeremy Kister
# Author: Jeremy Kister <esec - devel at jeremykister dot com>
# Date: 2003-Nov-01 22:32 (EDT)
# Name: Email Secretary (ESec)
# Function:  Challange Response system to deter spam

# loose mode lets user authenticate msg via reply-to or web
# in strict mode, user must load web page with random letters imbedded in random image displayed,

use strict;
use DBI;
use Net::SMTP;
use Net::DNS;

my $dbun = 'user';
my $dbpw = 'pass';
#my $dsn = 'DBI:Sybase:server=sql-01.example.net;database=esec'; # for MSSQL/Sybase
#my $dsn = 'DBI:mysql:host=sql-01.example.net;database=esec'; # for MySQL

my $account = shift || die "syntax error: no account specified\n";
my $version = '.26b';
my($maildir,$subject,$string,$noreply);
my(@msg,@header,@notes);
my $where = "pnd";
my $ext=$ENV{'EXT'} || die "environment not as expected: EXT\n";
my $host=$ENV{'HOST'} || die "environment not as expected: HOST\n";
my $sender=$ENV{'SENDER'}; # envelope sender: is empty string if bounce
my $email = "${ext}\@${host}";
if(($sender ne '') && ($sender !~ /^([a-z0-9_\.\+\-])+\@(([a-z0-9\-])+\.)+([a-z0-9]{2,4})+$/i)){
	warn "dropping message for ${account} (${email}): ${sender} doesnt look like a legit email address\n";
	exit;
}
my $rpline=$ENV{'RPLINE'};
my $dtline=$ENV{'DTLINE'};

open(M, '/var/qmail/control/me') || die "cannot open /var/qmail/control/me: $!\n";
chomp(my $hostname=<M>);
close M;

# have to read header before checking if user exists, to find out if
# we should bounce the message or not
#  protect ourselves from MTAs that send a message with no empty line
#  separating header and body: read only the first 200 lines, max
my $n=1;
while(($_=<STDIN>) && ($n < 201)){
	last if(/^$/); # we only want the headers, here.
	push @header, $_; # save header; we have no maildir, yet
	if(/^Subject\s*:\s?(.+)/i){
		$subject=$1;
	}elsif(/^X-ESec-Probe:\s(.{32})/){
		$string = $1;
		$noreply=1;
		push @notes, "string found in header: $string";
	}elsif(/^Precedence\s*:\s?(bulk|list|junk)/i){
		$noreply=1;
	}elsif(/^List-\S+/i){
		$noreply=1;
	}elsif(/^Mailing-List/i){
		$noreply=1;
	}
	$n++;
}

my $dbh = DBI->connect($dsn, $dbun, $dbpw,{RaiseError => 1});

# drop mail if person is listed in known_bouncers
my $sql = 'SELECT COUNT(*) FROM known_bouncers WHERE addr = ' . $dbh->quote($sender);
$sql .= " AND time < ($^T - 86400)"; # in case cron is not running for whatever reason
my $sth = $dbh->prepare($sql);
$sth->execute;
my $row=$sth->fetchrow_hashref;
my ($count) = values %$row;
exit if($count == 1);


my $sql = 'SELECT emailid,mode FROM accounts WHERE email = ' . $dbh->quote($account);
my $sth = $dbh->prepare($sql);
$sth->execute;
my $row = $sth->fetchrow_hashref;
$sth->finish;
my $emailid = $row->{emailid};
my $mode = $row->{mode};

if(defined($emailid)){
	if(my ($user,$domain) = split(/\@/, $account)){
		my $dir;
		open(A, "/var/qmail/users/assign") || die "cannot open assign: $!\n";
		while(<A>){
			if(/^\+${domain}\-:[^:]+:\d+:\d+:([^:]+):/){
				$dir = $1;
				last;
			}
		}
		close A;
		if(open(V, "${dir}/vpasswd")){
			while(<V>){
				if(/^${user}:[^:]+:\d+:\d+:[^:]*:([^:]+):/){
					$maildir = $1 . "/Maildir";
					last;
				}
			}
			close V;
			unless(defined($maildir)){
				exit if(defined($noreply));
				bounce('User Unknown #5.9.1',${hostname},$host,@header);
			}
		}else{
			warn "cannot open ${dir}/vpasswd: $!\n";
			exit if(defined($noreply));
			bounce('User Unknown #5.9.2',${hostname},$host,@header);
		}
	}else{
		warn "$account cant be split on an \@ (${email}) - why wasnt this msg dropped above?\n";
		exit if(defined($noreply));
		bounce('User Unknown #5.9.3',${hostname},$host,@header);
	}
}else{
	warn "no account set up for ${email}\n";
	exit if(defined($noreply));
	bounce('User Unknown #5.9.4',${hostname},$host,@header);
}

unless(-d "${maildir}/tmp"){
	unless(-d "${maildir}"){
		mkdir("${maildir}",0700) || die "cannot create ${maildir}: $!\n";
	}
	mkdir("${maildir}/tmp",0700) || die "cannot create ${maildir}/tmp: $!\n";
}
mkdir("${maildir}/pnd",0700) unless(-d "${maildir}/pnd");
mkdir("${maildir}/new",0700) unless(-d "${maildir}/new");

#stdin wound past the header
my $file="$^T.$$.esec.$hostname";
open(H,">${maildir}/tmp/${file}") || die "cannot write to ${maildir}/tmp/${file}: $!\n";
print H "$rpline",
        "$dtline";
foreach(@header){
	print H;
}
# we write to H below

# seperate file, because we need to know if body contains probe,
# but we are not done writing our header info
open(B, ">${maildir}/tmp/${file}.body") || die "cannot write to ${maildir}/tmp/${file}.body: $!\n";
my $n = 1;
my @body;
while(<STDIN>){
	print B;
	if($n < 21){
		# save first 20 lines of body in memory in case of a bounce
		push @body, $_;
		$n++;
	}
	if(/^X-ESec-Probe:\s(.{32})/){
		$string = $1;
		push @notes, "string found in body: $string";
	}
}
close B;

# bounces after this point must either read $file.body or have no body

if(defined($string)){
	# find out who this guy is
	my $sql = 'SELECT sender FROM messages WHERE msgid = ' . $dbh->quote($string);
	my $sth=$dbh->prepare($sql);
	$sth->execute;
	my $row=$sth->fetchrow_hashref;
	$sth->finish;
	my $dbsender = $row->{sender};

	if(defined($dbsender)){
		# this message is a bounce or a reply to my confirmation probe
		close B;
		close H;
		# unlink the tmp file
		unlink("${maildir}/tmp/${file}","${maildir}/tmp/${file}.body") || warn "cannot unlink tmp files: $!\n";
		
		if(($sender eq '') || ($sender =~ /^mailer-daemon\@/i)){
			# a bounce to my probe

			# add spammer to known bouncer's list, so we dont keep probing him
			my $sql = 'INSERT INTO known_bouncers (addr,time) VALUES (';
			$sql .= $dbh->quote($dbsender) . ",$^T)";
			my $sth=$dbh->prepare($sql);
			$sth->execute || die "cannot insert into known_bouncers: $!\n";

			# remove all messages in queue from this spammer (for this emailid only, not global)
			my $sql = 'SELECT filename FROM messages WHERE sender = ' . $dbh->quote($dbsender);
			$sql .= ' AND emailid = ' . $dbh->quote($emailid);
			my $sth = $dbh->prepare($sql);
			$sth->execute;
			while(my $row = $sth->fetchrow_hashref){
				my $filename = $row->{filename};
				unlink("${maildir}/pnd/${filename}") || warn "cannot unlink $maildir/pnd/$filename: $!\n";
			}

			my $sql = 'DELETE FROM messages WHERE sender = ' . $dbh->quote($dbsender);
			$sql .= ' AND emailid = ' . $dbh->quote($emailid);
			my $sth=$dbh->prepare($sql);
			$sth->execute || warn "could not delete msgs from $dbsender from db: $!\n";

		}elsif($sender eq $dbsender){
			if(($mode == 1) || ($sender eq $email)){
				# authenticate user
		
				# authenticate all messages from this sender (for this emailid)
				my $sql = 'SELECT filename FROM messages WHERE sender = ' . $dbh->quote($dbsender);
				$sql .= ' AND emailid = ' . $dbh->quote($emailid);
				my $sth = $dbh->prepare($sql);
				$sth->execute;
				while(my $row = $sth->fetchrow_hashref){
					my $filename = $row->{filename};
					if($sender eq $email){
						# from and to same address, do not accidentially let esec confirm the msg.
						# add $sender to known_bouncers
						unlink("${maildir}/pnd/${filename}") || warn "cannot unlink $maildir/pnd/$filename: $!\n";

						my $sql = 'INSERT INTO known_bouncers (addr,time) VALUES (';
						$sql .= $dbh->quote($sender) . ",$^T)";
						my $sth=$dbh->prepare($sql);
						$sth->execute || die "cannot insert into known_bouncers: $!\n";

					}elsif($mode == 1){
						rename("${maildir}/pnd/${filename}","${maildir}/new/${filename}") ||
					 	 warn "cannot rename $maildir/pnd/$filename: $!\n";
					}
				}
			
				my $sql = 'DELETE FROM messages WHERE sender = ' . $dbh->quote($dbsender);
				$sql .= ' AND emailid = ' . $dbh->quote($emailid);
				my $sth=$dbh->prepare($sql);
				$sth->execute || warn "could not delete msgs from $dbsender from db: $!\n";
			}else{
				exit if(defined($noreply));
				bounce('User Unknown #5.9.5',${hostname},$host,@header,@body);
			}
		}else{
			exit if(defined($noreply));
			bounce('User Unknown #5.9.6',${hostname},$host,@header,@body);
		}
		exit;
	}else{
		# or else the pending message using this string has been deleted from our queue
		unlink("${maildir}/tmp/${file}","${maildir}/tmp/${file}.body") || warn "cannot unlink tmp files: $!\n";
		exit;
	}
}
# spammers could easily foil this, but we'll let it in for now
#  there are rumors of laws against spammers forging the subject
if(($sender eq '') || ($sender =~ /^mailer-daemon\@/i)){
	$noreply=1;
	# let [most] bounces through
	if($subject eq 'failure notice'){
		$where = 'new';
		push @notes, 'Accepted because message looks like a qmail bounce';
	}elsif($subject =~ /^Undeliverable:\s\S/){
		$where = 'new';
		push @notes, 'Accepted because message looks like an exchange bounce';
	}elsif($subject =~ /^Undelivered Mail Returned to Sender/){
		$where = 'new';
		push @notes, 'Accepted because message looks like a postfix bounce';
	}elsif($subject =~ /^Returned mail:\s\S/){
		$where = 'new';
		push @notes, 'Accepted because message looks like a sendmail bounce';

	#Delivery Status Notification
	#Delivery Notification:\s\S
	#Delivery failure
	#Undeliverable Mail
	#Mail delivery failed:\s\S/
	}elsif($subject =~ /deliver|fail|status|notification/i){
		$where = 'new';
		push @notes, 'Accepted because message looks like a bounce';
	}
}

# could already be new because of a bounce
unless($where eq 'new'){
	# find out if the person is already known (via whitelist)
	my $sql = 'SELECT addr FROM email_whitelists WHERE emailid = ' . $dbh->quote($emailid);
	my $sth = $dbh->prepare($sql);
	$sth->execute;
	while(my $row = $sth->fetchrow_hashref){
		if($sender =~ /^$row->{addr}$/i){
			$where = "new";
			push @notes, "Accepted via email_whitelists: $row->{addr}";
			last;
		}
	}
	$sth->finish;

	unless($where eq 'new'){
		# find out if the email contains whitelisted headers
		my $sql = 'SELECT field FROM header_whitelists WHERE emailid = ' . $dbh->quote($emailid);
		my $sth = $dbh->prepare($sql);
		$sth->execute;
		while(my $row = $sth->fetchrow_hashref){
			foreach(@header){
				if(/^$row->{field}$/){
					$where = "new";
					push @notes, "Accepted via header_whitelists: $row->{field}";
					last;
				}
			}
		}
		$sth->finish;
	
		unless($where eq 'new'){
			# find out if the person is on our blacklist
			my $sql = 'SELECT addr FROM email_blacklists WHERE emailid = ' . $dbh->quote($emailid);
			my $sth = $dbh->prepare($sql);
			$sth->execute;
			while(my $row = $sth->fetchrow_hashref){
				if($sender =~ /^$row->{addr}$/i){
					# generic error, no red flags to sender
					unlink("${maildir}/tmp/${file}","${maildir}/tmp/${file}.body") || warn "cannot unlink tmp files: $!\n";
					exit if(defined($noreply));
					bounce('User Unknown #5.9.5',$hostname,$host,@header,@body);
				}
			}
			$sth->finish;

			my $sql = 'SELECT field FROM header_blacklists WHERE emailid = ' . $dbh->quote($emailid);
			my $sth = $dbh->prepare($sql);
			$sth->execute;
			while(my $row=$sth->fetchrow_hashref){
				foreach(@header){
					if(/^$row->{field}$/i){
						# generic error, no red flags to sender
						unlink("${maildir}/tmp/${file}","${maildir}/tmp/${file}.body") || warn "cannot unlink tmp files: $!\n";
						exit if(defined($noreply));
						bounce('User Unknown #5.9.7',$hostname,$host,@header,@body);
					}
				}
			}
			$sth->finish;
		}
	}
}

undef @header; # free a few bits of memory

# we dont want to send a confirmation on noreplies -- 
# and there'd be no point in saving the msg as pending
if(($where eq "pnd") && ($noreply == 1)){
	warn "msg for ${email} (from ${sender}) being dropped\n";
	close H;
	unlink("${maildir}/tmp/${file}","${maildir}/tmp/${file}.body") || warn "cannot unlink tmp files: $!\n";
	exit;
}

if(defined(@notes) && (${where} eq 'new')){
	print H 'X-ESec-Notes:';
	foreach my $note (@notes){
		print H " ${note}\n";
	}
}elsif(${where} eq 'pnd'){
	print H "X-ESec-Notes: Queued for confirmation\n";
}
my $date=getdate();
print H "X-ESec-Received: ${date}\n",
        "X-ESec-Version: ${version}\n",
	"\n"; #empty line separating header from body

open(B, "${maildir}/tmp/${file}.body") || die "cannot open ${maildir}/tmp/${file}.body: $!\n";
while(<B>){
	print H;
}
close B;
close H;
unlink("${maildir}/tmp/${file}.body") || warn "cannot unlink $maildir/tmp/$file.body: $!\n";

rename("${maildir}/tmp/${file}","${maildir}/${where}/${file}") || die "cannot rename $maildir/tmp/$file: $!\n";

exit if($where eq 'new'); # dont need to do any more 

my ($u,$d) = split(/\@/, $sender);
my (@mx) = mx($d);
foreach my $rr (@mx){
	my $exchanger = $rr->exchange;
	if(($exchanger =~ /^(127|0|10|255|224)\./) ||
	   ($exchanger =~ /^192\.168\./) ||
	   ($exchanger =~ /^172\.(1(6|7|8|9)|2(\d)|3(0|1))\./) || # stupid [16-31]
	   ($exchanger =~ /^192\.0\.2\./)){
		# bad mx record; delete pnd email
		# will always be pnd because of exit if where eq new above
		unlink("${maildir}/pnd/${file}") || warn "cannot unlink $maildir/pnd/$file: $!\n";
		exit;
	}
}

if(@mx){
	my($buf,$x,$random);
	open(D, "/dev/urandom") || die "cannot open urandom: $!\n";
	my @set = ('A'..'Z', 'a'..'z', '0'..'9', '_', '.');
	foreach(1..32){ # 256bit
		sysread( D, $buf, 1 );
		my $v = ord($buf);
		$x ^= ($v & ~63) >> (rand(7)+1);
		$random .= $set[ ($x ^ ord($buf)) & 63 ];
	}
	close D;

	# save the name of the file, who its from, and when we received it in the db
	my $sql = 'INSERT INTO messages (filename,sender,time,emailid,msgid) VALUES (' . $dbh->quote($file);
	$sql .= ',' . $dbh->quote($sender) . ',' . $dbh->quote($^T) . ',' . $dbh->quote($emailid);
	$sql .= ',' . $dbh->quote($random) . ')';
	my $sth = $dbh->prepare($sql);
	$sth->execute || die "error occurred saving info: $!\n";

	my $sql = 'SELECT COUNT(*) FROM messages WHERE sender = ' . $dbh->quote($sender);
	$sql .= ' AND msgid != ' . $dbh->quote($random) . " AND time > ($^T - 86400)";
	my $sth = $dbh->prepare($sql);
	$sth->execute;
	my $row = $sth->fetchrow_hashref;
	$sth->finish;
	my ($count) = values %$row;
	if($count > 0){

		# make sure we havent sent over 5 confirmations within 24 hours
		if($count > 5){
			# no confirmation msg, but keep message
			exit;
		}else{
			# make sure the last time we sent a confirmation to this person was over an hour ago
			my $sql = 'SELECT COUNT(*) FROM messages WHERE sender = ' . $dbh->quote($sender);
			$sql .= ' AND msgid != ' . $dbh->quote($random) . " AND time > ($^T - 3600)";
			my $sth = $dbh->prepare($sql);
			$sth->execute;
			my $row=$sth->fetchrow_hashref;
			$sth->finish;
			my ($recent) = values %$row;
			if($recent > 0){
				# no confirmation msg, but keep message
				exit;
			}
		}
	}

	# get the person's real name and unique message from the db
	my $sql = 'SELECT name,message FROM accounts WHERE emailid = ' . $dbh->quote($emailid);
	my $sth = $dbh->prepare($sql);
	$sth->execute;
	my $row=$sth->fetchrow_hashref;
	$sth->finish;
	$dbh->disconnect();
	my $name=$row->{name} || ${email};
	my $message=$row->{message};

	if(defined($message)){
		$message =~ s/%%name%%/${name}/g;
		$message =~ s/%%msgid%%/${random}/g;
		$message =~ s/%%subject%%/${subject}/g;
	}else{
		$message = "Hi,\n\n" .
			"This is the Email Secretary handling email for ${name}.  Because this is the first time I've seen an" .
			" email from this address, please confirm yourself so I will recognize you when you send in the future.\n\n" .
			"It's simple.  To prove your message came from a human (and not a spammer), go to:\n" .
			"http://qmailadmin.nntx.net/cgi-bin/eseccgi.pl?msgid=${random}\n\n";
		if($mode == 1){
			$message .= "You can also confirm yourself simply by simply replying to this message...\n" .
				"Just be sure to include this message in your reply.\n" .
				"X-ESec-Probe: ${random}\n\n";
		}
		$message .= "The subject of the email message you have sent (that is now in my Pending folder waiting for" .
			" your quick authentication) is: ${subject}\n\n" .
			"If you don't confirm this message within one week, I will delete it, and $name will never see it.\n\n";
	}
	open(Q, "|/var/qmail/bin/qmail-inject -f${email}") || die "cannot fork qmail-inject: $!\n";
	print Q "Date: ${date}\n",
		"Precedence: special-delivery\n",
		"To: <${sender}>\n",
		"From: \"${name}\" <${email}>\n",
	    "Subject: Re: $subject [::#:: ESec Confirmation ::#::]\n",
		"X-ESec-Probe: ${random}\n",
		"X-ESec-Version: ${version}\n",
		"\n",
		"${message}\n",
		"-----\n",
		"ESec: The Smarter Challenge-Response Anti-SPAM Solution.\n",
		"http://www.jeremykister.com/jeremy/code/perl/esec/\n";
	close Q;
}else{
	# no mx record; delete pnd email
	# where will always be pnd here, because "exit if new" above
	unlink("${maildir}/pnd/${file}") || die "cannot unlink $maildir/pnd/$file: $!\n"; 
}


sub bounce {
	my (($reason),$hostname,$host,@header,@body) = @_;

	# make sure domain has an mx record, and can connect on port 25
	# not everyone has dnsmx or dig, and dnsmx gives false answers ISO no mx
	my @mx = mx($host); # already sorted by preference
	foreach my $rr (@mx){
		my $exchanger = $rr->exchange;
		unless(($exchanger =~ /^(127|0|10|255|224)\./) ||
		       ($exchanger =~ /^192\.168\./) ||
		       ($exchanger =~ /^172\.(1(6|7|8|9)|2(\d)|3(0|1))\./) ||
		       ($exchanger =~ /^192\.0\.2\./)){
			my $smtp = Net::SMTP->new($exchanger);
			if(defined($smtp)){
				my $date = getdate();
				$smtp->mail("<>");
				$smtp->to("${sender}");
				$smtp->data();
				$smtp->datasend("Precedence: list\n");
				$smtp->datasend("Date: ${date}\n");
				$smtp->datasend("From: \"MAILER-DAEMON\@${hostname}\" <>\n");
				$smtp->datasend("To: ${sender}\n");
				$smtp->datasend("Subject: failure notice\n");
				$smtp->datasend("X-ESec-Version: ${version}\n\n");
				$smtp->datasend("Hi.  This is the ESec program at ${hostname}.\n");
				$smtp->datasend("I'm afraid I wasn't able to deliver your message to the following addresses.\n");
				$smtp->datasend("This is a permanent error; I've given up.  Sorry it didn't work out.\n\n");
				$smtp->datasend("<${email}>:\n");
			 	$smtp->datasend("${reason}\n\n");
				$smtp->datasend("--- Below this line is part of the message.\n\n");
				foreach(@header){
					$smtp->datasend($_);
				}
				$smtp->datasend("\n");
				if(defined(@body)){
					foreach(@body){
						$smtp->datasend($_);
						$n++;
					}
				}else{
					# show only first 20 lines of message
					my $n = 1;
					while( ($_=<STDIN>) && ($n < 21)){
						# wound past header
						$smtp->datasend($_);
						$n++;
					}
				}
				$smtp->dataend();
				exit;
			}
		} # or else bounce gets dumped
	} # or else bounce gets dumped
}

sub getdate {
	# user can set these up themselves 
	# http://atm.geo.nsf.gov/ieis/time.html
	my($day,$mon,$num,$time,$tz,$year) = split(/\s+/, `date`);
	my $offset="($tz)";
	if($tz eq 'EST'){
		$offset='-0500';
	}elsif($tz eq 'EDT'){
		$offset='-0400';
	}
	my $date = "${day}, ${num} ${mon} ${year} ${time} ${offset}";
	return($date);
}
